/*
 * file StpResource.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.StpResource
 *
 * (C) Copyright IBM Corporation 2004, 2011.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.wvcm.stp;


import static com.ibm.rational.wvcm.stpex.StpExBase.PROPERTY_NAMESPACE;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.List;

import javax.wvcm.Feedback;
import javax.wvcm.PropertyRequestItem;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;

import com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName;
import com.ibm.rational.wvcm.stpex.StpExBase;

/**
 * A proxy for a resource persisted in a repository or in a file area.
 */
public interface StpResource extends javax.wvcm.Resource
{
    /**
     * Returns true iff it can be determined that this proxy references the same
     * resource as the given proxy. Proxy equality is based either on the value
     * of the proxy's Location or on the value of the proxy's
     * RESOURCE_IDENTIFIER property.
     * <p>
     * This proxy equates to another proxy iff one of the following conditions
     * is true...
     * <ul>
     * <li>The location() of this proxy equals the location() of the other
     * proxy, OR
     * <li>The other proxy defines the RESOURCE_IDENTIFIER property and the
     * location().string() of this proxy equals the RESOURCE_IDENTIFIER of the
     * other proxy, OR
     * <li>This proxy defines the RESOURCE_IDENTIFIER property and it equals
     * the location().string() of the other proxy, OR
     * <li>Both proxies define the RESOURCE_IDENTIFIER property and they are
     * equal
     * </ul>
     * 
     * @param rhs The Object against which to compare this proxy. If the Object
     *            is not a Resource, equals returns <b>false</b>.
     * @return <b>true</b> if the proxies compare equal as defined above;
     *         <b>false</b> otherwise.
     * @see Object#equals(java.lang.Object)
     */
    boolean equals(Object rhs);

    /**
     * Returns the hashCode() for the value of this proxy's RESOURCE_IDENTIFIER
     * property if the proxy has a value for it; otherwise returns the
     * hashCode() for the proxy's Location object.
     * 
     * @see Object#hashCode()
     */
    int hashCode();

    /**
     * Returns the collection of metadata components defined in this proxy for a
     * given PropertyName.
     * 
     * @param name A PropertyName identifying the property whose metadata is
     *            desired
     * @return An StpProperty instance that contains the metadata for the given
     *         property requested from the server through this proxy.
     * @throws WvcmException if no metadata for the given property has been
     *             requested from the server through this proxy.
     */
    <T> StpProperty<T> getMetaProperties(PropertyName<T> name)
        throws WvcmException;

    /**
     * Request the properties specified in <code>feedback</code> from the
     * server resource represented by this proxy. A new proxy will be
     * constructed whose property map contains just the requested properties.
     * <p>
     * Properties are computed relative to the given context. For example, the
     * {@link com.ibm.rational.wvcm.stp.cc.CcVersion#VIEW_RELATIVE_PATH VIEW_RELATIVE_PATH}
     * property of a CcVersion or CcElement resource can only be provided in the
     * context of a CcView. Hence, you would use this form of doReadProperties
     * to specify the CcView context that should be used.
     * 
     * <p>
     * The properties requiring a CcView context are
     * 
     * <ul>
     * <li>{@link com.ibm.rational.wvcm.stp.cc.CcVersion#VIEW_RELATIVE_PATH}
     * <li>{@link com.ibm.rational.wvcm.stp.cc.CcElement#VIEW_RELATIVE_PATH}
     * <li>{@link com.ibm.rational.wvcm.stp.cc.CcFile#VIEW_RELATIVE_PATH}
     * </ul>
     * 
     * @param context An optional proxy (often CcView) providing context for the
     *            generation of certain properties (typically
     *            VIEW_RELATIVE_PATH) in the returned report. May be <b>null</b>.
     * @param feedback Provides optional feedback to the caller for long-running
     *            requests and/or a list of properties desired from the server.
     *            If the parameter is null or empty, a request is still made,
     *            and the returned proxy will have only the properties always
     *            requested from the server.
     * @return A proxy containing the properties retrieved from the server (and
     *         error messages for those that could not be retrieved). The class
     *         of the returned proxy is the one most suited to the type of
     *         resource addressed.
     * 
     * @throws WvcmException if failures occur
     * 
     * @see StpResource#doReadProperties(Feedback)
     */
    Resource doReadProperties(Resource context,
                              Feedback feedback) throws WvcmException;

    /**
     * Returns whether or not from this proxy uncorrupted property values can be
     * obtained for all the properties and meta-properties specified by a given
     * PropertyRequestItem, which may specify multiple and nested properties.
     * <p>
     * If <code>wantedProp</code> is a nested property request then the test
     * is applied recursively to the non-null value(s) of the property named by
     * <code>wantedProp.getRoot()</code> testing in each value for valid
     * instances of the properties named by <code>wantedProp.getNested()</code>.
     * If any of the values identified by the property name are invalid, then
     * this method returns <code>false</code>. Note that if the value for the
     * root property of a PropertyName is <code>null</code>, the test for
     * that PropertyName is deemed valid even if the PropertyName had nested
     * properties.
     * 
     * @param wantedProp A PropertyRequestItem, possibly with multiple nested
     *            properties, to check the value for. Cannot be
     *            <code>null</code>.
     * 
     * @return <code>true</code> if there is an entry in this proxy for all
     *         properties named at the top level of the PropertyRequestItem, and
     *         those entries contain a valid value (not in error) and for those
     *         property values that are resources any nested property request is
     *         also similarly satisfied; <code>false</code> otherwise.
     */
    boolean hasProperties(PropertyRequestItem wantedProp);

    /**
     * Sets the value of the specified property in this resource proxy.
     * This method differs from setProperty() in that this method does not mark
     * the property "dirty", so subsequent calls to doWriteProperties() or other
     * "do()" methods will not commit this new value to the underlying resource.
     * @param name the name of the property.
     * @param value the new value of the specified property.
     * @see #setProperty
     * @see #getProperty
     */
    public <T> void setPropertyClean(PropertyName<T> name, T value);

    /**
     * Returns the StpPropertyException that would be thrown by getProperty if
     * it were called with the given PropertyName.
     * 
     * <p>
     * If there was an error retrieving the property when requested, or the
     * property was never defined in the proxy, an {@link StpPropertyException}
     * is returned; otherwise <b>null</b> is returned.
     * 
     * @param propertyName The property for which an exception is desired.
     *            Cannot be <b>null</b>.
     * 
     * @return The exception that would be thrown by getProperty() if it were
     *         used to access the given <code>propertyName</code>; Will be
     *         <b> null</b> if no exception would be thrown.
     */
    StpException getPropertyException(PropertyName<?> propertyName);

    /**
     * The most human-readable location for a resource. This is the form of
     * location that clients should show to users when a precise location must
     * be displayed--e.g. one which the user must remember and re-enter in a
     * subsequent session. When such precision is not needed, the
     * {@link javax.wvcm.Resource#DISPLAY_NAME} property would be a better
     * choice for the presentation.
     * <p>
     * The USER_FRIENDLY_LOCATION is <i>usually</i> expressed by a
     * user-friendly-selector scheme location, but other location schemes may
     * also be used depending on the type of resource.
     * <p>
     * The value of this property is <i>not</i> always defined for every
     * resource. If a resource does not have a USER_FRIENDLY_LOCATION, that is a
     * strong indication that that resource should not be part of the normal
     * object model presented to the user. If not defined, the value will be an
     * empty string.
     * <p>
     * Note that the location used by a proxy returned by the server is
     * formatted to provide the most efficient access to the resource in
     * subsequent interactions with the server in the current session. That
     * location is not necessarily the same as the resource's
     * USER_FRIENDLY_LOCATION or its STABLE_LOCATION.
     * 
     * @see #STABLE_LOCATION
     * @see #EFFICIENT_LOCATION
     */
    PropertyName<StpLocation> USER_FRIENDLY_LOCATION =
        new PropertyName<StpLocation>(PROPERTY_NAMESPACE,
                                      "user-friendly-location");

    /**
     * Returns the value of the
     * {@link #USER_FRIENDLY_LOCATION USER_FRIENDLY_LOCATION} property as
     * defined by this proxy.
     * 
     * @return The StpLocation value. Will never be <b>null</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #USER_FRIENDLY_LOCATION USER_FRIENDLY_LOCATION}
     *             property.
     */
    StpLocation getUserFriendlyLocation() throws WvcmException;

    /**
     * The most stable (over space and time) location available for this
     * resource. Clients should use this form of location when archiving
     * resource references for use between sessions.
     * <p>
     * The stable location is <i>usually</i> expressed by a repo-selector
     * scheme location, but other location schemes may also be used depending on
     * the type of resource.
     * <p>
     * The value of this property is always defined for every resource, but, in
     * fact, <i>it may not be stable</i>. But, if the resource has a stable
     * location, that location will be returned by this property.
     * <p>
     * The value returned always represents the identifier associated with the
     * server state of the resource, even if the Resource on which it is
     * requested has client-side state.
     * <p>
     * Note that the location used by a proxy returned by the server is
     * formatted to provide the <i>most efficient</i> access to the resource in
     * subsequent interactions with the server in the current session. That
     * location is not necessarily the same as the resource's
     * USER_FRIENDLY_LOCATION or its STABLE_LOCATION.
     * 
     * @see #USER_FRIENDLY_LOCATION
     * @see #EFFICIENT_LOCATION
     */
    PropertyName<StpLocation> STABLE_LOCATION =
        new PropertyName<StpLocation>(PROPERTY_NAMESPACE, "stable-location");

    /**
     * Returns the value of the {@link #STABLE_LOCATION STABLE_LOCATION}
     * property as defined by this proxy.
     * 
     * @return The StpLocation value. Will never be <b>null</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #STABLE_LOCATION STABLE_LOCATION} property.
     */
    StpLocation getStableLocation() throws WvcmException;

    /**
     * The most efficient location available for this resource. Clients should
     * use this form of location when it is not necessary to either show this
     * location to users, nor to archive resource references for use between
     * sessions.
     * <p>
     * The efficient location is <i>usually</i> expressed by a repo-selector
     * scheme location, but other location schemes may also be used depending on
     * the type of resource.
     * <p>
     * The value of this property is always defined for every resource, but, in
     * fact, <i>it may not be more efficient than the other forms</i>. But, if
     * the resource has an efficient location, that location will be returned by
     * this property.
     * <p>
     * The value returned always represents the identifier associated with the
     * server state of the resource, even if the Resource on which it is
     * requested has client-side state.
     * <p>
     * This is the location format usually used in the proxies returned from the
     * server.
     * 
     * @see #USER_FRIENDLY_LOCATION
     * @see #STABLE_LOCATION
     */
    PropertyName<StpLocation> EFFICIENT_LOCATION =
        new PropertyName<StpLocation>(PROPERTY_NAMESPACE, "efficient-location");

    /**
     * Returns the value of the {@link #EFFICIENT_LOCATION EFFICIENT_LOCATION}
     * property as defined by this proxy.
     * 
     * @return The StpLocation value for the resources most efficient location.
     *         Will never be <b>null</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #EFFICIENT_LOCATION EFFICIENT_LOCATION} property.
     */
    StpLocation getEfficientLocation() throws WvcmException;

    /**
     * Note: the {@link #RESOURCE_IDENTIFIER} property is defined only for
     * server resources (it's undefined for uncontrolled controllable resources)
     * and it's value is not available locally via
     * ControllableResource.readProperties().
     * 
     * @see javax.wvcm.Resource#getResourceIdentifier()
     * @see javax.wvcm.Resource#RESOURCE_IDENTIFIER
     */
    public String getResourceIdentifier() throws WvcmException;

    /**
     * The authentication realm in which this resource resides. When credentials
     * are needed to access this resource this is string that should be
     * presented to the user to identify what the user is to provide credentials
     * for.
     */
    PropertyName<String> AUTHENTICATION_REALM =
        new PropertyName<String>(PROPERTY_NAMESPACE, "authentication-realm");

    /**
     * Returns the value of the {@link #AUTHENTICATION_REALM
     * AUTHENTICATION_REALM} property as defined by this proxy.
     * 
     * @return A String containing the realm identifier passed to the
     *         authentication callback when accessing this resource.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #AUTHENTICATION_REALM AUTHENTICATION_REALM} property.
     */
    String getAuthenticationRealm() throws WvcmException;

    /** A list of the properties that are currently invalid for this resource */
    PropertyName<StpProperty.List<StpProperty<?>>> INVALID_PROPERTIES =
        new PropertyName<StpProperty.List<StpProperty<?>>>(PROPERTY_NAMESPACE,
                                                           "invalid-properties");

    /**
     * Returns the value of the {@link #INVALID_PROPERTIES INVALID_PROPERTIES}
     * property as defined by this proxy.
     * 
     * @return A Property.List of Property objects, each identifying one invalid
     *         property of this Resource
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #INVALID_PROPERTIES INVALID_PROPERTIES} property.
     */
    StpProperty.List<StpProperty<?>> getInvalidProperties()
        throws WvcmException;

    /**
     * A list of all the properties that are defined on this resource. The
     * returned properties can be retrieved as a list using getAllProperties()
     * or individually, using
     * {@link javax.wvcm.Resource#getProperty(javax.wvcm.PropertyNameList.PropertyName) getProperty} or an
     * appropriate getter method.
     */
    PropertyName<StpProperty.List<StpProperty<?>>> ALL_PROPERTIES =
        new PropertyName<StpProperty.List<StpProperty<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                           "all-properties");

    /**
     * Returns the properties defined in this proxy. Note, this "getter" will
     * work even if ALL_PROPERTIES was not explicitly requested when building
     * the proxy.
     * 
     * @return An StpProperty.List containing a Property object for each
     *         property defined by this proxy. If there were errors in
     *         retrieving any meta-properties (including the VALUE
     *         meta-property), these errors can be discovered through the
     *         StpProperty interface.
     */
    StpProperty.List<StpProperty<?>> getAllProperties();

    /**
     * Returns the custom properties recorded in this proxy. Note, this "getter"
     * will work even if ALL_CUSTOM_PROPERTIES was not explicitly requested when
     * building the proxy.
     * 
     * @return A Property.List containing a Property object for each custom
     *         (i.e. not defined by WVCM) property defined by this proxy. If
     *         there were errors in retrieving any meta-properties (including
     *         the VALUE meta-property), these errors can be discovered through
     *         the Property interface.
     */
    StpProperty.List<StpProperty<?>> getCustomProperties();

    /** The StpRepository that contains this resource. */
    PropertyName<StpRepository> REPOSITORY =
        new PropertyName<StpRepository>(PROPERTY_NAMESPACE, "repository");

    /**
     * Returns the value of the {@link #REPOSITORY} property as defined by this
     * proxy.
     * 
     * @return The StpRepository that this resource is in. Will never be <b>null</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #REPOSITORY} property.
     */
    StpRepository getRepository() throws WvcmException;

    /**
     * The login name of the user that created this object. A request for this
     * value can only be satisfied by a request to a server; it is not available
     * locally.
     */
    PropertyName<String> CREATOR_LOGIN_NAME =
        new PropertyName<String>(PROPERTY_NAMESPACE, "creator-login-name");

    /**
     * Returns the value of the {@link #CREATOR_LOGIN_NAME} property as defined
     * by this proxy.
     * 
     * @throws WvcmException Thrown if the property was never requested, or
     *             otherwise unavailable (e.g. requested in a client-context
     *             only).
     */
    String getCreatorLoginName() throws WvcmException;

    /**
     * The group name of the user that created an object. This value can only be
     * satisfied by a request to the server; it is not available locally.
     */
    PropertyName<String> CREATOR_GROUP_NAME =
        new PropertyName<String>(PROPERTY_NAMESPACE, "creator-group-name");

    /**
     * Returns the value of the {@link #CREATOR_GROUP_NAME} property as defined
     * by this proxy.
     * 
     * @throws WvcmException Thrown if the property was never requested, or is
     *             otherwise unavailable (e.g. requested in a client-context
     *             only).
     */
    String getCreatorGroupName() throws WvcmException;

    /**
     * Returns the Location of this proxy's Location cast as an StpLocation.
     * 
     * @return The StpLocation for this Resource proxy. Will never be <b>null</b>.
     */
    StpLocation stpLocation();

    /**
     * Returns the value of the <i>resource error attribute</i> associated with
     * this proxy.
     * 
     * <p>
     * If this attribute is not null it indicates that an error occurred when
     * trying to access the resource addressed by this proxy (even though that
     * address was actually generated by the server). In this case, none of the
     * properties requested for this proxy will be defined. The value of the
     * resource error attribute is an StpException that records the nature of
     * the problem encountered by the server.
     * 
     * <p>
     * This attribute is set only when this proxy is returned as an indirect
     * result of a server interaction. If this defective address was used by a
     * client as the target of a server contact method or if it was the expected
     * direct result of some server contact method, that contact method would
     * have thrown an exception rather than setting this attribute.
     * 
     * <p>
     * Whenever the server returns a list of resources (either as a direct
     * response to a request or as the value of a property) it is possible that
     * the server can enumerate all of the members of the list but cannot
     * actually access one or more of the enumerated members to access its
     * properties. When this happens, the operation is not aborted and does not
     * itself throw an exception. Instead, a proxy is generated and marked with
     * this resource error attribute so that the problem is record and then the
     * rest of the list is processed.
     * 
     * @return <b>null</b> if no error is associated with the resource,
     *         otherwise a StpException describing the error.
     */
    StpException getResourceError();

    /**
     * @return Returns the StpProvider from which this proxy was obtained.
     */
    StpProvider stpProvider();

    /**
     * Sets the value of a meta-property of a property into this proxy without
     * causing that property to be written to the resource on the next
     * repository interaction. Like the setProperty() method, this method does
     * not verify that the given object is a value suitable for the specified
     * meta-property. This method also does not verify that the given
     * PropertyName is appropriate for the type of resource represented by this
     * proxy nor that the given MetaPropertyName is appropriate for the given
     * PropertyName.
     * <p>
     * This method may be used in conjunction with
     * {@link StpProvider#buildProxy(StpLocation, String, String) StpProvider.buildProxy()}
     * to reconstruct proxies previously saved to persistent storage. Use
     * buildProxy() to construct a proxy object of the appropriate type and then
     * use this method to copy each saved meta-property value into that proxy
     * object.
     * <p>
     * This method may also be used in conjunction with forgetProperty to merge
     * meta-property values from one proxy into another and thereby facilitate
     * the use of a proxy object as a long-lived cache for resource data. This
     * approach is not recommended unless the client fully understands the
     * pitfalls inherent in maintaining a cache of shared server data in a
     * client for any length of time.
     * <p style="padding-left:47;text-indent:-47">
     * Note 1: Since the primary intent of this method is to provide the client
     * with a means to restore a proxy from archive, this method will throw an
     * exception if an attempt is made to define a meta-property value already
     * being defined by this proxy.
     * <p style="padding-left:47">
     * If a client wants to use this method to update an existing meta-property
     * entry within a proxy, that existing meta-property value must first be
     * removed. Since the meta-properties of a property are usually tightly
     * coupled, they cannot be removed independently. They can only be removed
     * as a group using the javax.wvcm.forgetProperty() method. If the client
     * wants to retain some of the existing meta-property values while updating
     * others, it must first fetch all the existing meta-property values from
     * the proxy (using getMetaProperties()), then remove their definitions from
     * the proxy using forgetProperty(), and then redefine the individual
     * meta-property entries via multiple invocations of this method, reusing
     * the old values or computing new values in a manner appropriate to the
     * client's needs.
     * <p style="padding-left:47;text-indent:-47">
     * Note 2: Because of its role in the definition of Resource.equals() and
     * Resource.hashCode(), the Resource.RESOURCE_IDENTIFIER property may only
     * be initialized at the time of proxy construction; e.g., via
     * Provider.buildProxy().
     * <p style="padding-left:47;text-indent:-47">
     * Note 3: A property value set by any of the setXXX methods defined by this
     * interface (or any of its extensions) will activate the writing of that
     * value to the repository by the next doXYZ method. Use of initProperty to
     * set any of the meta-properties of such a property will fail until the
     * property is successfully written to the repository.
     * <p style="padding-left:47;text-indent:-47">
     * Note 4: If the entry in this proxy for the given PropertyName indicates
     * that an error occurred on the last attempt to read the property from the
     * repository, initMetaProperty will fail as if the meta-property being set
     * were already defined; i.e. it will throw an exception. In this case, the
     * client must execute forgetProperty() on the invalid entry before
     * attempting to initialize any of its meta-properties.
     * 
     * @param propertyName The PropertyName for the property whose meta-property
     *            is to be defined by the proxy. Must not be <b>null</b>, must
     *            not be equal to Resource.RESOURCE_IDENTIFIER, and must not be
     *            a PropertyName currently on the
     *            {@link javax.wvcm.Resource#updatedPropertyNameList()}.
     * @param metaPropertyName The MetaPropertyName for the meta-property whose
     *            value is to be defined by the proxy. Must not be <b>null</b>
     *            and must not name a meta-property whose value is already
     *            defined by this proxy.
     * @param value The initial value for the meta-property identified by the
     *            other two arguments of this method. May be <b>null</b>.
     * @throws WvcmException If the meta-property is already defined by this
     *             proxy.
     * 
     * @see StpProvider#buildProxy(StpLocation, String, String)
     * @see javax.wvcm.Resource#forgetProperty(javax.wvcm.PropertyNameList.PropertyName)
     */
    <T> void initMetaProperty(PropertyName<?> propertyName,
                              MetaPropertyName<T> metaPropertyName,
                              T value) throws WvcmException;

    /**
     * Returns an identifier that may be passed to StpProvider.buildProxy() to
     * construct a proxy that supports the same interfaces as this proxy.
     * 
     * @return A String object containing the proxy class information; will
     *         never be <b>null</b>.
     */
    String proxyType();

    /**
     * An annotation used on property getters or setters of a proxy interface to
     * indicate that a property defined in a super interface is not supported by
     * that proxy.
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.METHOD)
    public @interface UnsupportedProperty {
        /**
         * @return The StpReasonCode of the StpException that will be thrown if
         *         the annotated method is used.
         */
        StpException.StpReasonCode value();
    }
}
